grammar DLV;

options {
	output=AST;
	ASTLabelType=CustomANTLRASTNode;
	superClass=AbstractCustomANTLRParser;
}


// These are imaginary tokens that will serve as parent nodes
// for grouping other tokens in our AST.
tokens {
  PROGRAM;
  INTERPRETATION;
  BLOCK;
  RULE;
  STATEMENT;
  HEAD;
  BODY;
  FACT;
  CONSTRAINT;
  WEAKCONSTRAINT;
  LEVEL;
  WEIGHT;
  NAF;
  NEG;
  ATOM;
  BUILDINLITERAL;
  PLUS;
  TIMES;
  INTRANGEATOM;
  TERMS;
  SIMPLECOMMENT;
  META;
  PROPERTY;
  ERROR;
//query stuff
  QAND;
  QOR;
  QEXISTS;
  QBEXISTS;
  QFORALL;
  QBFORALL;
}

@parser::header{
	package org.mmdasp.solver.dlv.parsing;
    import org.mmdasp.code.ast.IRuleNode;
}

@parser::members {		

}

@lexer::header{
	package org.mmdasp.solver.dlv.parsing;
	import java.util.HashMap;
}

@lexer::members {
	private boolean commentMode=false;
	private boolean ignoreNewLines=true;		
	private int offset=0;
	HashMap<Integer,Integer> lineStartOffsets = new HashMap<Integer,Integer>();
	
	public void initialise() {
		lineStartOffsets.put(new Integer(1) , new Integer(0));
	}
	
	public HashMap<Integer,Integer> getLineStartOffsets(){
		return lineStartOffsets;
	}
}


// first order query parsing ////

programAndQuery :		program query							->		program query;

queryEOF :				query EOF								->		query; 

query :					'?' queryBody '.'						->		queryBody; 

queryBody :				qOrAnd 									->		qOrAnd
					 |	qQuant									->		qQuant;

qQuant :				IDLOWER
						{$IDLOWER.text.equals("exists")||
						 $IDLOWER.text.equals("forall")}?
						IDUPPER (',' IDUPPER)* '.' queryBody 	->		{$IDLOWER.text.equals("exists")}?
																			^(QEXISTS IDUPPER+ queryBody)
																-> 			^(QFORALL IDUPPER+ queryBody);

qOrAnd :			 	qAndQuery
						(IDLOWER {$IDLOWER.text.equals("v")}?
						 qAndQuery)*							->		^(QOR qAndQuery+);

qAndQuery :				qDirectConjuncts (',' qDirectConjuncts)*->		^(QAND qDirectConjuncts+);

qDirectConjuncts :		classicalLiteral						->		classicalLiteral
					|	builtinPredicate						->		builtinPredicate
					|	qPar									->		qPar
					|	qNotQuery								->		qNotQuery;

qNotQuery :				NOT qDirectConjuncts				 	->		^(NAF qDirectConjuncts);


qPar :					'(' queryBody ')'	 					->		queryBody;



/////////////////////


programAndInterpretation :
						program interpretation					->		program interpretation;


programEOF :			program EOF								->		program;

program :				rulestatorblock* comment*				->		^(PROGRAM rulestatorblock* comment*);

interpretationEOF :		interpretation EOF						->		interpretation;


interpretation :		LBRACE (classicalLiteral
						 (',' classicalLiteral)*)? RBRACE		->		^(INTERPRETATION classicalLiteral*);

rulestatorblock :		rule									->		rule
					|	statement								->		statement	
					|	block									->		block;

statement :				comment* maxintdecl						->		^(STATEMENT ^(maxintdecl comment*)) ;

maxintdecl :			SPECIALMAXINT
						comment* EQUAL comment* NUM
						comment* '.'							->		^(SPECIALMAXINT NUM comment*);


block :					comment*
						blockbegin
						rulestatorblock*
						comment*
						blockend								->		^(BLOCK rulestatorblock* comment*);


rule :					comment* standardrule					->		^(standardrule comment*)
					|	comment* fact 							->		^(fact comment*)
					|	comment* constraint						->		^(constraint comment*)
					|	comment* weakConstraint 				->		^(weakConstraint comment*);

standardrule :			head ':-' body '.'						->		^(RULE head body);


fact :	
						head ':-'? '.'							->		^(FACT head)
					|   integerRangeAtom '.'					->		^(FACT integerRangeAtom);

constraint :			':-' body '.'							->		^(CONSTRAINT body);

weakConstraint :		':~' body '.' (comment* weightLevel)?	->		^(WEAKCONSTRAINT body weightLevel? comment*);

weightLevel :			'[' weight ':' level ']'				->		weight level
					|	'[' weight ':'       ']'				->		weight
					|	'['        ':' level ']'				->		level;

weight:					comment* IDUPPER comment*				->		^(WEIGHT IDUPPER comment*)
					|	comment* NUM comment*					->		^(WEIGHT NUM comment*); 

level:					comment* IDUPPER comment*				->		^(LEVEL IDUPPER comment*)
					|	comment* NUM comment*					->		^(LEVEL NUM comment*); 
// erzeuge fehler wenn WS fehlt - noetig? ! 
head : 					classicalLiteral
						(((IDLOWER {$IDLOWER.text.equals("v")}?)|'|') 
						classicalLiteralLC)*					->		^(HEAD classicalLiteral classicalLiteralLC*);
				

body : 					literalLC (',' literalLC)*				->		^(BODY literalLC+);

literalLC :				comment* literal 						->		^(literal comment*);


literal :		 		classicalLiteral 						->		classicalLiteral 
					|	NOT classicalLiteralLC					->		^(NAF classicalLiteralLC)
					|	builtinPredicate 						->		builtinPredicate;

classicalLiteralLC :	comment* classicalLiteral				->		^(classicalLiteral comment*);

classicalLiteral :		atom 									->		atom
					|	CNOT atomLC								->		^(NEG atomLC);

atomLC:					comment* atom							->		^(atom comment*);

atom :		 			predicate comment*						->		^(ATOM predicate comment*)
					|	predicate comment* 
					'(' (termLC (',' termLC )*)? ')' comment*	->		^(ATOM predicate ^(TERMS termLC*) comment*);
		
		
integerRangeAtom :		predicate comment* 
						'(' numTermLC
						DOTDOT numTermLC  ')' comment*			->		^(INTRANGEATOM predicate ^(TERMS numTermLC*) comment*);
					
builtinPredicate : 		comparativePredicate					->		^(BUILDINLITERAL comparativePredicate)
					|	arithmeticPredicate						->		^(BUILDINLITERAL arithmeticPredicate);

comparativePredicate :	term comparisonOperator termLC			->		^(comparisonOperator ^(TERMS term termLC))
					|	comparisonOperator comment*
						'(' termLC ',' termLC ')' comment*		->		^(comparisonOperator ^(TERMS termLC termLC) comment*);
					
comparisonOperator :	DOUBLEEQUAL								->		EQUAL
					|	EQUAL									->		EQUAL
					|	LESS									->		LESS
					|	LESSEQUAL								->		LESSEQUAL
					|	GREATER									->		GREATER
					|	GREATEREQUAL							->		GREATEREQUAL
					|	NOTEQUAL								->		NOTEQUAL;


arithmeticPredicate :	specialIntPredicate						->		specialIntPredicate
					|	specialSuccPredicate					->		specialSuccPredicate
					|	term EQUAL termLC '+' termLC			->		^(PLUS ^(TERMS termLC termLC term))
					|	term EQUAL termLC '*' termLC			->		^(TIMES ^(TERMS termLC termLC term))
					|	'+' comment* '(' termLC ',' termLC','
						termLC ')' comment*						->		^(PLUS ^(TERMS termLC*) comment*)
					|	'*' comment* '(' termLC ',' termLC','
						termLC ')' comment*						->		^(TIMES ^(TERMS termLC*) comment*);

specialIntPredicate :	SPECIALINT
						comment* '(' termLC ')' comment*		->		^(SPECIALINT ^(TERMS termLC) comment*);

specialSuccPredicate :	SPECIALSUCC comment*
						'(' termLC ',' termLC ')' comment*		->		^(SPECIALSUCC ^(TERMS termLC*) comment*);

numTermLC :				comment* numeric comment*				->		^(numeric comment*);

termLC :				comment* term							->		^(term comment*);

term : 					constant comment*						->		^(constant comment*)
					|	SPECIALMAXINT comment*					->		^(SPECIALMAXINT comment*)
					|	variable comment*						->		^(variable comment*);

// eigenes PRED schafft Probleme -> wieso blos?
predicate :				id										->		id;


numeric :				NUM										->		NUM
					|	SPECIALMAXINT							->		SPECIALMAXINT;

constant :				IDLOWER									->		IDLOWER
					|	NUM										->		NUM;

variable : 				IDUPPER									->		IDUPPER
					|	ANOVAR									->		ANOVAR;
					
comment :				EMPTYCOMMENT							->
					|	STARTMETACOMM metacontent?
						(NEWLINE|EOF)							->		^(META metacontent?)
					|	IGNORECOMMENT							->		^(SIMPLECOMMENT IGNORECOMMENT);

metacontent :			(metastatement ';')*
						metastatement (';')?					->		metastatement+;

metastatement :			id '=' id?								->		^(PROPERTY id id?);

blockbegin	:			STARTMETACOMM 'begin block' (NEWLINE | EOF);
blockend	:			STARTMETACOMM 'end block' (NEWLINE | EOF);

id:						IDLOWER									->		^(IDLOWER)
					|	IDUPPER									->		^(IDUPPER);						


EMPTYCOMMENT	:		COMMENTSYMBOL (NEWLINE | EOF);
STARTMETACOMM	:		COMMENTSYMBOL BANG {ignoreNewLines=false;commentMode=true;};



IGNORECOMMENT : 		COMMENTSYMBOL {ignoreNewLines=false;commentMode=true;} (~( BANG |'\r\n' | '\r' | '\n' )) (~('\r\n' | '\r' | '\n' ))* (NEWLINE | EOF);


fragment
COMMENTSYMBOL :			'%' ;

fragment
BANG :					'!';
				
NUM:					('0'..'9')+ ;


//V : 					{!commentMode}? 'v';

NOT :					{!commentMode}? 'not';

DOTDOT :				{!commentMode}? '..';

EQUAL	:				{!commentMode}? '=';

DOUBLEEQUAL :			{!commentMode}? '==';

NOTEQUAL :	 			{!commentMode}? '!='
					|	{!commentMode}? '<>';


LESS :	 				{!commentMode}? '<';
LESSEQUAL :	 			{!commentMode}? '<=';

GREATER :	 			{!commentMode}? '>';
GREATEREQUAL :			{!commentMode}? '>=';

SPECIALMAXINT : 		{!commentMode}? '#maxint';

SPECIALINT :			{!commentMode}? '#int';
SPECIALSUCC :			{!commentMode}? '#succ';

SPECIALCOUNT :			{!commentMode}? '#count';
SPECIALSUM :			{!commentMode}? '#sum';
SPECIALTIMES :			{!commentMode}? '#times';
SPECIALMIN :			{!commentMode}? '#min';
SPECIALMAX :			{!commentMode}? '#max';

SPECIALEXPORT :			{!commentMode}? '#export';
SPECIALIMPORT :			{!commentMode}? '#import';

SPECIALCONST :			{!commentMode}? '#const';

CNOT :					{!commentMode}? '-' | '~' ;

LBRACE :				'{';

RBRACE :				'}';

IDUPPER : 				'A'..'Z' ('a'..'z' | 'A'..'Z' | '0'..'9'| '_')* ;

ANOVAR : 				'_' ;

// check with semantic predicate that it does not spell "not"
IDLOWER : 				'a'..'z' ('a'..'z' | 'A'..'Z' | '0'..'9'| '_')* ;

fragment
WS : 					(NEWLINE |WSSAMELINE) ;

WSSAMELINE :			(' ' |'\t'|'\f') {$channel=HIDDEN;};

NEWLINE 
@after{if(!ignoreNewLines)ignoreNewLines=true;commentMode=false;}
:				(NEWLINE1 | NEWLINE2) {if(ignoreNewLines){ skip();}};

fragment
NEWLINE1 :				('\r' | '\n' ) {lineStartOffsets.put(new Integer(state.tokenStartLine+1), new Integer(((ICustomANTLRInputStream)input).getCurrentPosition()));};

// for windows systems
fragment
NEWLINE2 :				'\r\n' {lineStartOffsets.put(new Integer(state.tokenStartLine+1), new Integer(((ICustomANTLRInputStream)input).getCurrentPosition()));};
